"""
Featurizer implementations used in Smiles2Vec models.
SmilesToSeq featurizer for Smiles2Vec models taken from https://arxiv.org/abs/1712.02734

"""
from typing import Dict, List
import numpy as np
import pandas as pd

from deepchem.utils.typing import RDKitMol
from deepchem.feat.base_classes import MolecularFeaturizer

PAD_TOKEN = "<pad>"
OUT_OF_VOCAB_TOKEN = "<unk>"


def create_char_to_idx(filename: str,
                       max_len: int = 250,
                       smiles_field: str = "smiles") -> Dict[str, int]:
    """Creates a dictionary with character to index mapping.

    Parameters
    ----------
    filename: str
        Name of the file containing the SMILES strings
    max_len: int, default 250
        Maximum allowed length of the SMILES string
    smiles_field: str, default "smiles"
        Field indicating the SMILES strings int the file.

    Returns
    -------
    Dict[str, int]
        A dictionary mapping characters to their integer indexes.

    """
    smiles_df = pd.read_csv(filename)
    char_set = set()
    for smile in smiles_df[smiles_field]:
        if len(smile) <= max_len:
            char_set.update(set(smile))

    unique_char_list = list(char_set)
    unique_char_list += [PAD_TOKEN, OUT_OF_VOCAB_TOKEN]
    char_to_idx = {letter: idx for idx, letter in enumerate(unique_char_list)}
    return char_to_idx


class SmilesToSeq(MolecularFeaturizer):
    """
    SmilesToSeq Featurizer takes a SMILES string, and turns it into a sequence.
    Details taken from [1]_.

    SMILES strings smaller than a specified max length (max_len) are padded using
    the PAD token while those larger than the max length are not considered. Based
    on the paper, there is also the option to add extra padding (pad_len) on both
    sides of the string after length normalization. Using a character to index (char_to_idx)
    mapping, the SMILES characters are turned into indices and the
    resulting sequence of indices serves as the input for an embedding layer.

    References
    ----------
    .. [1] Goh, Garrett B., et al. "Using rule-based labels for weak supervised
        learning: a ChemNet for transferable chemical property prediction."
        Proceedings of the 24th ACM SIGKDD International Conference on Knowledge
        Discovery & Data Mining. 2018.

    Note
    ----
    This class requires RDKit to be installed.

    """

    def __init__(self,
                 char_to_idx: Dict[str, int],
                 max_len: int = 250,
                 pad_len: int = 10):
        """Initialize this class.

        Parameters
        ----------
        char_to_idx: Dict
            Dictionary containing character to index mappings for unique characters
        max_len: int, default 250
            Maximum allowed length of the SMILES string.
        pad_len: int, default 10
            Amount of padding to add on either side of the SMILES seq

        """
        self.max_len = max_len
        self.char_to_idx = char_to_idx
        self.idx_to_char = {
            idx: letter for letter, idx in self.char_to_idx.items()
        }
        self.pad_len = pad_len

    def to_seq(self, smile: List[str]) -> np.ndarray:
        """Turns list of smiles characters into array of indices"""
        out_of_vocab_idx = self.char_to_idx[OUT_OF_VOCAB_TOKEN]
        seq = [
            self.char_to_idx.get(character, out_of_vocab_idx)
            for character in smile
        ]
        return np.array(seq)

    def remove_pad(self, characters: List[str]) -> List[str]:
        """Removes PAD_TOKEN from the character list."""
        characters = characters[self.pad_len:len(characters) - self.pad_len]
        chars = list()

        for char in characters:
            if char != PAD_TOKEN:
                chars.append(char)
        return chars

    def smiles_from_seq(self, seq: List[int]) -> str:
        """Reconstructs SMILES string from sequence."""
        characters = [self.idx_to_char[i] for i in seq]

        characters = self.remove_pad(characters)
        smile = "".join([letter for letter in characters])
        return smile

    def _featurize(self, datapoint: RDKitMol, **kwargs) -> np.ndarray:
        """Featurizes a SMILES sequence.

        Parameters
        ----------
        datapoints: rdkit.Chem.rdchem.Mol
            RDKit Mol object

        Returns
        -------
        np.ndarray
            A 1D array of a SMILES sequence.
            If the length of SMILES is longer than `max_len`, this value is an empty array.

        """
        try:
            from rdkit import Chem
        except ModuleNotFoundError:
            raise ImportError("This class requires RDKit to be installed.")

        if 'mol' in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )
        smile = Chem.MolToSmiles(datapoint)
        if len(smile) > self.max_len:
            return np.array([])

        smile_list = list(smile)
        # Extend shorter strings with padding
        if len(smile) < self.max_len:
            smile_list.extend([PAD_TOKEN] * (self.max_len - len(smile)))

        # Padding before and after
        smile_list += [PAD_TOKEN] * self.pad_len
        smile_list = [PAD_TOKEN] * self.pad_len + smile_list

        smile_seq = self.to_seq(smile_list)
        return smile_seq



"""
DGL-based PAGTN for graph property prediction.
"""
import torch.nn as nn
import torch.nn.functional as F

from deepchem.models.losses import Loss, L2Loss, SparseSoftmaxCrossEntropy
from deepchem.models.torch_models.torch_model import TorchModel


class Pagtn(nn.Module):
    """Model for Graph Property Prediction

    This model proceeds as follows:

    * Update node representations in graphs with a variant of GAT, where a
        linear additive form of attention is applied. Attention Weights are derived
        by concatenating the node and edge features for each bond.
    * Update node representations with multiple rounds of message passing.
    * For each layer has, residual connections with its previous layer.
    * The final molecular representation is computed by combining the representations
        of all nodes in the molecule.
    * Perform the final prediction using a linear layer

    Examples
    --------

    >>> import deepchem as dc
    >>> import dgl
    >>> from deepchem.models import Pagtn
    >>> smiles = ["C1CCC1", "C1=CC=CN=C1"]
    >>> featurizer = dc.feat.PagtnMolGraphFeaturizer(max_length=5)
    >>> graphs = featurizer.featurize(smiles)
    >>> print(type(graphs[0]))
    <class 'deepchem.feat.graph_data.GraphData'>
    >>> dgl_graphs = [graphs[i].to_dgl_graph() for i in range(len(graphs))]
    >>> batch_dgl_graph = dgl.batch(dgl_graphs)
    >>> model = Pagtn(n_tasks=1, mode='regression')
    >>> preds = model(batch_dgl_graph)
    >>> print(type(preds))
    <class 'torch.Tensor'>
    >>> preds.shape == (2, 1)
    True

    References
    ----------
    .. [1] Benson Chen, Regina Barzilay, Tommi Jaakkola. "Path-Augmented
        Graph Transformer Network." arXiv:1905.12712

    Notes
    -----
    This class requires DGL (https://github.com/dmlc/dgl) and DGL-LifeSci
    (https://github.com/awslabs/dgl-lifesci) to be installed.
    """

    def __init__(self,
                 n_tasks: int,
                 number_atom_features: int = 94,
                 number_bond_features: int = 42,
                 mode: str = 'regression',
                 n_classes: int = 2,
                 output_node_features: int = 256,
                 hidden_features: int = 32,
                 num_layers: int = 5,
                 num_heads: int = 1,
                 dropout: float = 0.1,
                 nfeat_name: str = 'x',
                 efeat_name: str = 'edge_attr',
                 pool_mode: str = 'sum'):
        """
        Parameters
        ----------
        n_tasks: int
            Number of tasks.
        number_atom_features : int
            Size for the input node features. Default to 94.
        number_bond_features : int
            Size for the input edge features. Default to 42.
        mode: str
            The model type, 'classification' or 'regression'. Default to 'regression'.
        n_classes: int
            The number of classes to predict per task
            (only used when ``mode`` is 'classification'). Default to 2.
        output_node_features : int
            Size for the output node features in PAGTN layers. Default to 256.
        hidden_features : int
            Size for the hidden node features in PAGTN layers. Default to 32.
        num_layers : int
            Number of PAGTN layers to be applied. Default to 5.
        num_heads : int
            Number of attention heads. Default to 1.
        dropout : float
            The probability for performing dropout. Default to 0.1
        nfeat_name: str
            For an input graph ``g``, the model assumes that it stores node features in
            ``g.ndata[nfeat_name]`` and will retrieve input node features from that.
            Default to 'x'.
        efeat_name: str
            For an input graph ``g``, the model assumes that it stores edge features in
            ``g.edata[efeat_name]`` and will retrieve input edge features from that.
            Default to 'edge_attr'.
        pool_mode : 'max' or 'mean' or 'sum'
            Whether to compute elementwise maximum, mean or sum of the node representations.
        """
        try:
            import dgl  # noqa: F401
        except:
            raise ImportError('This class requires dgl.')
        try:
            import dgllife  # noqa: F401
        except:
            raise ImportError('This class requires dgllife.')

        if mode not in ['classification', 'regression']:
            raise ValueError(
                "mode must be either 'classification' or 'regression'")

        super(Pagtn, self).__init__()

        self.n_tasks = n_tasks
        self.mode = mode
        self.n_classes = n_classes
        self.nfeat_name = nfeat_name
        self.efeat_name = efeat_name
        if mode == 'classification':
            out_size = n_tasks * n_classes
        else:
            out_size = n_tasks

        from dgllife.model import PAGTNPredictor as DGLPAGTNPredictor

        self.model = DGLPAGTNPredictor(node_in_feats=number_atom_features,
                                       node_out_feats=output_node_features,
                                       node_hid_feats=hidden_features,
                                       edge_feats=number_bond_features,
                                       depth=num_layers,
                                       nheads=num_heads,
                                       dropout=dropout,
                                       n_tasks=out_size,
                                       mode=pool_mode)

    def forward(self, g):
        """Predict graph labels

        Parameters
        ----------
        g: DGLGraph
            A DGLGraph for a batch of graphs. It stores the node features in
            ``dgl_graph.ndata[self.nfeat_name]`` and edge features in
            ``dgl_graph.edata[self.efeat_name]``.

        Returns
        -------
        torch.Tensor
            The model output.

        * When self.mode = 'regression',
            its shape will be ``(dgl_graph.batch_size, self.n_tasks)``.
        * When self.mode = 'classification', the output consists of probabilities
            for classes. Its shape will be
            ``(dgl_graph.batch_size, self.n_tasks, self.n_classes)`` if self.n_tasks > 1;
            its shape will be ``(dgl_graph.batch_size, self.n_classes)`` if self.n_tasks is 1.
        torch.Tensor, optional
            This is only returned when self.mode = 'classification', the output consists of the
            logits for classes before softmax.
        """
        node_feats = g.ndata[self.nfeat_name]
        edge_feats = g.edata[self.efeat_name]
        out = self.model(g, node_feats, edge_feats)

        if self.mode == 'classification':
            if self.n_tasks == 1:
                logits = out.view(-1, self.n_classes)
                softmax_dim = 1
            else:
                logits = out.view(-1, self.n_tasks, self.n_classes)
                softmax_dim = 2
            proba = F.softmax(logits, dim=softmax_dim)
            return proba, logits
        else:
            return out


class PagtnModel(TorchModel):
    """Model for Graph Property Prediction.

    This model proceeds as follows:

    * Update node representations in graphs with a variant of GAT, where a
        linear additive form of attention is applied. Attention Weights are derived
        by concatenating the node and edge features for each bond.
    * Update node representations with multiple rounds of message passing.
    * For each layer has, residual connections with its previous layer.
    * The final molecular representation is computed by combining the representations
        of all nodes in the molecule.
    * Perform the final prediction using a linear layer

    Examples
    --------
    >>> import deepchem as dc
    >>> from deepchem.models import PagtnModel
    >>> # preparing dataset
    >>> smiles = ["C1CCC1", "CCC"]
    >>> labels = [0., 1.]
    >>> featurizer = dc.feat.PagtnMolGraphFeaturizer(max_length=5)
    >>> X = featurizer.featurize(smiles)
    >>> dataset = dc.data.NumpyDataset(X=X, y=labels)
    >>> # training model
    >>> model = PagtnModel(mode='classification', n_tasks=1,
    ...                    batch_size=16, learning_rate=0.001)
    >>> loss = model.fit(dataset, nb_epoch=5)

    References
    ----------
    .. [1] Benson Chen, Regina Barzilay, Tommi Jaakkola. "Path-Augmented
        Graph Transformer Network." arXiv:1905.12712

    Notes
    -----
    This class requires DGL (https://github.com/dmlc/dgl) and DGL-LifeSci
    (https://github.com/awslabs/dgl-lifesci) to be installed.
    """

    def __init__(self,
                 n_tasks: int,
                 number_atom_features: int = 94,
                 number_bond_features: int = 42,
                 mode: str = 'regression',
                 n_classes: int = 2,
                 output_node_features: int = 256,
                 hidden_features: int = 32,
                 num_layers: int = 5,
                 num_heads: int = 1,
                 dropout: float = 0.1,
                 pool_mode: str = 'sum',
                 **kwargs):
        """
        Parameters
        ----------
        n_tasks: int
            Number of tasks.
        number_atom_features : int
            Size for the input node features. Default to 94.
        number_bond_features : int
            Size for the input edge features. Default to 42.
        mode: str
            The model type, 'classification' or 'regression'. Default to 'regression'.
        n_classes: int
            The number of classes to predict per task
            (only used when ``mode`` is 'classification'). Default to 2.
        output_node_features : int
            Size for the output node features in PAGTN layers. Default to 256.
        hidden_features : int
            Size for the hidden node features in PAGTN layers. Default to 32.
        num_layers: int
            Number of graph neural network layers, i.e. number of rounds of message passing.
            Default to 2.
        num_heads : int
            Number of attention heads. Default to 1.
        dropout: float
            Dropout probability. Default to 0.1
        pool_mode : 'max' or 'mean' or 'sum'
            Whether to compute elementwise maximum, mean or sum of the node representations.
        kwargs
            This can include any keyword argument of TorchModel.
        """
        model = Pagtn(n_tasks=n_tasks,
                      number_atom_features=number_atom_features,
                      number_bond_features=number_bond_features,
                      mode=mode,
                      n_classes=n_classes,
                      output_node_features=output_node_features,
                      hidden_features=hidden_features,
                      num_layers=num_layers,
                      num_heads=num_heads,
                      dropout=dropout,
                      pool_mode=pool_mode)
        if mode == 'regression':
            loss: Loss = L2Loss()
            output_types = ['prediction']
        else:
            loss = SparseSoftmaxCrossEntropy()
            output_types = ['prediction', 'loss']
        super(PagtnModel, self).__init__(model,
                                         loss=loss,
                                         output_types=output_types,
                                         **kwargs)

    def _prepare_batch(self, batch):
        """Create batch data for Pagtn.

        Parameters
        ----------
        batch: tuple
            The tuple is ``(inputs, labels, weights)``.

        Returns
        -------
        inputs: DGLGraph
            DGLGraph for a batch of graphs.
        labels: list of torch.Tensor or None
            The graph labels.
        weights: list of torch.Tensor or None
            The weights for each sample or sample/task pair converted to torch.Tensor.
        """
        try:
            import dgl
        except:
            raise ImportError('This class requires dgl.')

        inputs, labels, weights = batch
        dgl_graphs = [graph.to_dgl_graph() for graph in inputs[0]]
        inputs = dgl.batch(dgl_graphs).to(self.device)
        _, labels, weights = super(PagtnModel, self)._prepare_batch(
            ([], labels, weights))
        return inputs, labels, weights



import numpy as np
from rdkit import Chem
from rdkit.Chem import AllChem, rdMolAlign

from deepchem.feat.graph_data import GraphData
from deepchem.feat import MolecularFeaturizer

# similar to SNAP featurizer. both taken from Open Graph Benchmark (OGB) github.com/snap-stanford/ogb
# The difference between this and the SNAP features is the lack of masking tokens, possible_implicit_valence_list, possible_bond_dirs
# and the prescence of possible_bond_stereo_list,  possible_is_conjugated_list, possible_is_in_ring_list,
allowable_features = {
    'possible_atomic_num_list': list(range(1, 119)) + ['misc'],  # type: ignore
    'possible_chirality_list': [
        'CHI_UNSPECIFIED', 'CHI_TETRAHEDRAL_CW', 'CHI_TETRAHEDRAL_CCW',
        'CHI_OTHER', 'misc'
    ],
    'possible_degree_list': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 'misc'],
    'possible_formal_charge_list': [
        -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 'misc'
    ],
    'possible_numH_list': [0, 1, 2, 3, 4, 5, 6, 7, 8, 'misc'],
    'possible_number_radical_e_list': [0, 1, 2, 3, 4, 'misc'],
    'possible_hybridization_list': [
        'SP', 'SP2', 'SP3', 'SP3D', 'SP3D2', 'misc'
    ],
    'possible_is_aromatic_list': [False, True],
    'possible_is_in_ring_list': [False, True],
    'possible_bond_type_list': [
        'SINGLE', 'DOUBLE', 'TRIPLE', 'AROMATIC', 'misc'
    ],
    'possible_bond_stereo_list': [
        'STEREONONE',
        'STEREOZ',
        'STEREOE',
        'STEREOCIS',
        'STEREOTRANS',
        'STEREOANY',
    ],
    'possible_is_conjugated_list': [False, True],
}

full_atom_feature_dims = list(
    map(
        len,  # type: ignore
        [
            allowable_features['possible_atomic_num_list'],
            allowable_features['possible_chirality_list'],
            allowable_features['possible_degree_list'],
            allowable_features['possible_formal_charge_list'],
            allowable_features['possible_numH_list'],
            allowable_features['possible_number_radical_e_list'],
            allowable_features['possible_hybridization_list'],
            allowable_features['possible_is_aromatic_list'],
            allowable_features['possible_is_in_ring_list']
        ]))

full_bond_feature_dims = list(
    map(
        len,  # type: ignore
        [
            allowable_features['possible_bond_type_list'],
            allowable_features['possible_bond_stereo_list'],
            allowable_features['possible_is_conjugated_list']
        ]))


def safe_index(feature_list, e):
    """
    Return index of element e in list l. If e is not present, return the last index

    Parameters
    ----------
    feature_list : list
        Feature vector
    e : int
        Element index to find in feature vector
    """
    try:
        return feature_list.index(e)
    except ValueError:
        return len(feature_list) - 1


class RDKitConformerFeaturizer(MolecularFeaturizer):
    """
    A featurizer that featurizes an RDKit mol object as a GraphData object with 3D coordinates. The 3D coordinates are represented in the node_pos_features attribute of the GraphData object of shape [num_atoms * num_conformers, 3].

    The ETKDGv2 algorithm is used to generate 3D coordinates for the molecule.
    The RDKit source for this algorithm can be found in RDkit/Code/GraphMol/DistGeomHelpers/Embedder.cpp
    The documentation can be found here:
    https://rdkit.org/docs/source/rdkit.Chem.rdDistGeom.html#rdkit.Chem.rdDistGeom.ETKDGv2

    This featurization requires RDKit.

    Examples
    --------
    >>> from deepchem.feat.molecule_featurizers.conformer_featurizer import RDKitConformerFeaturizer
    >>> from deepchem.feat.graph_data import BatchGraphData
    >>> import numpy as np
    >>> featurizer = RDKitConformerFeaturizer(num_conformers=2)
    >>> molecule = "CCO"
    >>> features_list = featurizer.featurize([molecule])
    >>> batched_feats = BatchGraphData(np.concatenate(features_list).ravel())
    >>> print(batched_feats.node_pos_features.shape)
    (18, 3)
    """

    def __init__(self, num_conformers: int = 1, rmsd_cutoff: float = 2):
        """
        Initialize the RDKitConformerFeaturizer with the given parameters.

        Parameters
        ----------
        num_conformers : int, optional, default=1
            The number of conformers to generate for each molecule.
        rmsd_cutoff : float, optional, default=2
            The root-mean-square deviation (RMSD) cutoff value. Conformers with an RMSD
            greater than this value will be discarded.
        """
        self.num_conformers = num_conformers
        self.rmsd_cutoff = rmsd_cutoff

    def atom_to_feature_vector(self, atom):
        """
        Converts an RDKit atom object to a feature list of indices.

        Parameters
        ----------
        atom : Chem.rdchem.Atom
            RDKit atom object.

        Returns
        -------
        List[int]
            List of feature indices for the given atom.
        """
        atom_feature = [
            safe_index(allowable_features['possible_atomic_num_list'],
                       atom.GetAtomicNum()),
            safe_index(allowable_features['possible_chirality_list'],
                       str(atom.GetChiralTag())),
            safe_index(allowable_features['possible_degree_list'],
                       atom.GetTotalDegree()),
            safe_index(allowable_features['possible_formal_charge_list'],
                       atom.GetFormalCharge()),
            safe_index(allowable_features['possible_numH_list'],
                       atom.GetTotalNumHs()),
            safe_index(allowable_features['possible_number_radical_e_list'],
                       atom.GetNumRadicalElectrons()),
            safe_index(allowable_features['possible_hybridization_list'],
                       str(atom.GetHybridization())),
            allowable_features['possible_is_aromatic_list'].index(
                atom.GetIsAromatic()),
            allowable_features['possible_is_in_ring_list'].index(
                atom.IsInRing()),
        ]
        return atom_feature

    def bond_to_feature_vector(self, bond):
        """
        Converts an RDKit bond object to a feature list of indices.

        Parameters
        ----------
        bond : Chem.rdchem.Bond
            RDKit bond object.

        Returns
        -------
        List[int]
            List of feature indices for the given bond.
        """
        bond_feature = [
            safe_index(allowable_features['possible_bond_type_list'],
                       str(bond.GetBondType())),
            allowable_features['possible_bond_stereo_list'].index(
                str(bond.GetStereo())),
            allowable_features['possible_is_conjugated_list'].index(
                bond.GetIsConjugated()),
        ]
        return bond_feature

    def _featurize(self, datapoint):
        """
        Featurizes a molecule into a graph representation with 3D coordinates.

        Parameters
        ----------
        datapoint : RdkitMol
            RDKit molecule object

        Returns
        -------
        graph: List[GraphData]
            list of GraphData objects of the molecule conformers with 3D coordinates.
        """
        # add hydrogen bonds to molecule because they are not in the smiles representation
        mol = Chem.AddHs(datapoint)
        try:
            ps = AllChem.ETKDGv2()
            ps.useRandomCoords = True
            AllChem.EmbedMolecule(mol, ps)
            AllChem.EmbedMultipleConfs(mol, self.num_conformers)
            AllChem.MMFFOptimizeMolecule(mol)
            rmsd_list = []
            rdMolAlign.AlignMolConformers(mol, RMSlist=rmsd_list)
            # insert 0 RMSD for first conformer
            rmsd_list.insert(0, 0)
            conformers = [
                mol.GetConformer(i)
                for i in range(self.num_conformers)
                if rmsd_list[i] < self.rmsd_cutoff
            ]
            # if conformer list is less than num_conformers, pad by repeating conformers
            conf_idx = 0
            while len(conformers) < self.num_conformers:
                conformers.append(conformers[conf_idx])
                conf_idx += 1

            coordinates = [conf.GetPositions() for conf in conformers]
        except Exception as e:
            print(e)
            print("Error featurizing molecule: ", datapoint)

        atom_features_list = []
        for atom in mol.GetAtoms():
            atom_features_list.append(self.atom_to_feature_vector(atom))

        edges_list = []
        edge_features_list = []
        for bond in mol.GetBonds():
            i = bond.GetBeginAtomIdx()
            j = bond.GetEndAtomIdx()
            edge_feature = self.bond_to_feature_vector(bond)

            # add edges in both directions
            edges_list.append((i, j))
            edge_features_list.append(edge_feature)
            edges_list.append((j, i))
            edge_features_list.append(edge_feature)

        # Graph connectivity in COO format with shape [2, num_edges]
        graph_list = []
        for i in range(self.num_conformers):
            graph_list.append(
                GraphData(node_pos_features=np.array(coordinates[i]),
                          node_features=np.array(atom_features_list),
                          edge_features=np.array(edge_features_list),
                          edge_index=np.array(edges_list).T))
        return graph_list



"""
Contains an abstract base class that supports chemically aware data splits.
"""
import numpy as np
from deepchem.data import NumpyDataset
from deepchem.splits import Splitter


def merge_fold_datasets(fold_datasets):
    """Merges fold datasets together.

    Assumes that fold_datasets were outputted from k_fold_split. Specifically,
    assumes that each dataset contains the same datapoints, listed in the same
    ordering.
    """
    if not len(fold_datasets):
        return None

    # All datasets share features and identifiers by assumption.
    X = fold_datasets[0].X
    ids = fold_datasets[0].ids

    ys, ws = [], []
    for fold_dataset in fold_datasets:
        ys.append(fold_dataset.y)
        ws.append(fold_dataset.w)
    y = np.concatenate(ys, axis=1)
    w = np.concatenate(ws, axis=1)
    return NumpyDataset(X, y, w, ids)


class TaskSplitter(Splitter):
    """Provides a simple interface for splitting datasets task-wise.

    For some learning problems, the training and test datasets should
    have different tasks entirely. This is a different paradigm from the
    usual Splitter, which ensures that split datasets have different
    datapoints, not different tasks.
    """

    def __init__(self):
        "Creates Task Splitter object."
        pass

    def train_valid_test_split(self,
                               dataset,
                               frac_train=.8,
                               frac_valid=.1,
                               frac_test=.1):
        """Performs a train/valid/test split of the tasks for dataset.

        If split is uneven, spillover goes to test.

        Parameters
        ----------
        dataset: dc.data.Dataset
            Dataset to be split
        frac_train: float, optional
            Proportion of tasks to be put into train. Rounded to nearest int.
        frac_valid: float, optional
            Proportion of tasks to be put into valid. Rounded to nearest int.
        frac_test: float, optional
            Proportion of tasks to be put into test. Rounded to nearest int.
        """
        np.testing.assert_almost_equal(frac_train + frac_valid + frac_test, 1)
        n_tasks = len(dataset.get_task_names())
        n_train = int(np.round(frac_train * n_tasks))
        n_valid = int(np.round(frac_valid * n_tasks))

        X, y, w, ids = dataset.X, dataset.y, dataset.w, dataset.ids

        train_dataset = NumpyDataset(X, y[:, :n_train], w[:, :n_train], ids)
        valid_dataset = NumpyDataset(X, y[:, n_train:n_train + n_valid],
                                     w[:, n_train:n_train + n_valid], ids)
        test_dataset = NumpyDataset(X, y[:, n_train + n_valid:],
                                    w[:, n_train + n_valid:], ids)
        return train_dataset, valid_dataset, test_dataset

    def k_fold_split(self, dataset, K):
        """Performs a K-fold split of the tasks for dataset.

        If split is uneven, spillover goes to last fold.

        Parameters
        ----------
        dataset: dc.data.Dataset
            Dataset to be split
        K: int
            Number of splits to be made
        """
        n_tasks = len(dataset.get_task_names())
        n_per_fold = int(np.round(n_tasks / float(K)))
        if K * n_per_fold != n_tasks:
            print("Assigning extra tasks to last fold due to uneven split")

        X, y, w, ids = dataset.X, dataset.y, dataset.w, dataset.ids

        fold_datasets = []
        for fold in range(K):
            if fold != K - 1:
                fold_tasks = range(fold * n_per_fold, (fold + 1) * n_per_fold)
            else:
                fold_tasks = range(fold * n_per_fold, n_tasks)
            if len(w.shape) == 1:
                w_tasks = w
            elif w.shape[1] == 1:
                w_tasks = w[:, 0]
            else:
                w_tasks = w[:, fold_tasks]
            fold_datasets.append(NumpyDataset(X, y[:, fold_tasks], w_tasks,
                                              ids))
        return fold_datasets



import deepchem as dc
import numpy as np
import pytest
try:
    import torch
    has_torch = True
except ModuleNotFoundError:
    has_torch = False


@pytest.mark.torch
def test_unsorted_segment_sum():

    segment_ids = torch.Tensor([0, 1, 0]).to(torch.int64)
    data = torch.Tensor([[1, 2, 3, 4], [5, 6, 7, 8], [4, 3, 2, 1]])
    num_segments = 2

    # length of segment_ids.shape should be 1
    assert len(segment_ids.shape) == 1

    # Shape of segment_ids should be equal to first dimension of data
    assert segment_ids.shape[-1] == data.shape[0]
    result = dc.utils.pytorch_utils.unsorted_segment_sum(
        data=data, segment_ids=segment_ids, num_segments=num_segments)

    assert np.allclose(
        np.array(result),
        np.load("deepchem/utils/test/assets/result_segment_sum.npy"),
        atol=1e-04)



import numpy as np
import logging
from deepchem.feat import MaterialStructureFeaturizer
from collections import defaultdict
from typing import List, Dict, Tuple, DefaultDict, Any
from deepchem.utils.typing import PymatgenStructure
from deepchem.feat.graph_data import GraphData
from scipy.spatial.distance import pdist, squareform, cdist
from scipy.spatial.transform import Rotation


class LCNNFeaturizer(MaterialStructureFeaturizer):
    """
    Calculates the 2-D Surface graph features in 6 different permutations-

    Based on the implementation of Lattice Graph Convolution Neural
    Network (LCNN). This method produces the Atom wise features ( One Hot Encoding)
    and Adjacent neighbour in the specified order of permutations. Neighbors are
    determined by first extracting a site local environment from the primitive cell,
    and perform graph matching and distance matching to find neighbors.
    First, the template of the Primitive cell needs to be defined along with periodic
    boundary conditions and active and spectator site details. structure(Data Point
    i.e different configuration of adsorbate atoms) is passed for featurization.

    This particular featurization produces a regular-graph (equal number of Neighbors)
    along with its permutation in 6 symmetric axis. This transformation can be
    applied when orderering of neighboring of nodes around a site play an important role
    in the propert predictions. Due to consideration of local neighbor environment,
    this current implementation would be fruitful in finding neighbors for calculating
    formation energy of adbsorption tasks where the local. Adsorption turns out to be important
    in many applications such as catalyst and semiconductor design.

    The permuted neighbors are calculated using the Primitive cells i.e periodic cells
    in all the data points are built via lattice transformation of the primitive cell.

    `Primitive cell Format:`

    1. Pymatgen structure object with site_properties key value
        - "SiteTypes" mentioning if it is a active site "A1" or spectator
            site "S1".
    2. ns , the number of spectator types elements. For "S1" its 1.
    3. na , the number of active types elements. For "A1" its 1.
    4. aos, the different species of active elements "A1".
    5. pbc, the periodic boundary conditions.

    `Data point Structure Format(Configuration of Atoms):`

    1. Pymatgen structure object with site_properties with following key value.
        - "SiteTypes", mentioning if it is a active site "A1" or spectator
            site "S1".
        - "oss", different occupational sites. For spectator sites make it -1.

    It is highly recommended that cells of data are directly redefined from
    the primitive cell, specifically, the relative coordinates between sites
    are consistent so that the lattice is non-deviated.

    References
    ----------
    .. [1] Jonathan Lym and Geun Ho Gu, J. Phys. Chem. C 2019, 123, 18951−18959

    Examples
    --------
    >>> import deepchem as dc
    >>> from pymatgen.core import Structure
    >>> import numpy as np
    >>> PRIMITIVE_CELL = {
    ...   "lattice": [[2.818528, 0.0, 0.0],
    ...               [-1.409264, 2.440917, 0.0],
    ...               [0.0, 0.0, 25.508255]],
    ...   "coords": [[0.66667, 0.33333, 0.090221],
    ...              [0.33333, 0.66667, 0.18043936],
    ...              [0.0, 0.0, 0.27065772],
    ...              [0.66667, 0.33333, 0.36087608],
    ...              [0.33333, 0.66667, 0.45109444],
    ...              [0.0, 0.0, 0.49656991]],
    ...   "species": ['H', 'H', 'H', 'H', 'H', 'He'],
    ...   "site_properties": {'SiteTypes': ['S1', 'S1', 'S1', 'S1', 'S1', 'A1']}
    ... }
    >>> PRIMITIVE_CELL_INF0 = {
    ...    "cutoff": np.around(6.00),
    ...    "structure": Structure(**PRIMITIVE_CELL),
    ...    "aos": ['1', '0', '2'],
    ...    "pbc": [True, True, False],
    ...    "ns": 1,
    ...    "na": 1
    ... }
    >>> DATA_POINT = {
    ...   "lattice": [[1.409264, -2.440917, 0.0],
    ...               [4.227792, 2.440917, 0.0],
    ...               [0.0, 0.0, 23.17559]],
    ...   "coords": [[0.0, 0.0, 0.099299],
    ...              [0.0, 0.33333, 0.198598],
    ...              [0.5, 0.16667, 0.297897],
    ...              [0.0, 0.0, 0.397196],
    ...              [0.0, 0.33333, 0.496495],
    ...              [0.5, 0.5, 0.099299],
    ...              [0.5, 0.83333, 0.198598],
    ...              [0.0, 0.66667, 0.297897],
    ...              [0.5, 0.5, 0.397196],
    ...              [0.5, 0.83333, 0.496495],
    ...              [0.0, 0.66667, 0.54654766],
    ...              [0.5, 0.16667, 0.54654766]],
    ...   "species": ['H', 'H', 'H', 'H', 'H', 'H',
    ...               'H', 'H', 'H', 'H', 'He', 'He'],
    ...   "site_properties": {
    ...     "SiteTypes": ['S1', 'S1', 'S1', 'S1', 'S1',
    ...                   'S1', 'S1', 'S1', 'S1', 'S1',
    ...                   'A1', 'A1'],
    ...     "oss": ['-1', '-1', '-1', '-1', '-1', '-1',
    ...             '-1', '-1', '-1', '-1', '0', '2']
    ...                   }
    ... }
    >>> featuriser = dc.feat.LCNNFeaturizer(**PRIMITIVE_CELL_INF0)
    >>> print(type(featuriser._featurize(Structure(**DATA_POINT))))
    <class 'deepchem.feat.graph_data.GraphData'>

    Notes
    -----
    This Class requires pymatgen , networkx , scipy installed.
    """

    def __init__(self,
                 structure: PymatgenStructure,
                 aos: List[str],
                 pbc: List[bool],
                 ns: int = 1,
                 na: int = 1,
                 cutoff: float = 6.00):
        """
        Parameters
        ----------
        structure: : PymatgenStructure
            Pymatgen Structure object of the primitive cell used for calculating
            neighbors from lattice transformations.It also requires site_properties
            attribute with "Sitetypes"(Active or spectator site).
        aos: List[str]
            A list of all the active site species. For the Pt, N, NO configuration
            set it as ['0', '1', '2']
        pbc: List[bool]
            Periodic Boundary Condition
        ns: int (default 1)
            The number of spectator types elements. For "S1" its 1.
        na: int (default 1)
            the number of active types elements. For "A1" its 1.
        cutoff: float (default 6.00)
            Cutoff of radius for getting local environment.Only
            used down to 2 digits.
        """
        try:
            from pymatgen.core import Structure
        except:
            raise ImportError("This class requires pymatgen to be installed.")

        if type(structure) is not Structure:
            structure = Structure(**structure)
        self.aos = aos
        self.cutoff = np.around(cutoff, 2)
        self.setup_env = _load_primitive_cell(structure, aos, pbc, ns, na,
                                              cutoff)

    def _featurize(self, datapoint: PymatgenStructure, **kwargs) -> GraphData:
        """
        Parameters
        ----------
        datapoint: : PymatgenStructure
            Pymatgen Structure object of the surface configuration. It also requires
            site_properties attribute with "Sitetypes"(Active or spectator site) and
            "oss"(Species of Active site from the list of self.aos and "-1" for
            spectator sites).

        Returns
        -------
        graph: GraphData
            Node features, All edges for each node in diffrent permutations
        """
        if 'structure' in kwargs and datapoint is None:
            datapoint = kwargs.get("structure")
            raise DeprecationWarning(
                'Structure is being phased out as a parameter, please pass "datapoint" instead.'
            )

        xSites, xNSs = self.setup_env.read_datum(datapoint)
        config_size = xNSs.shape
        v = np.arange(0, len(xSites)).repeat(config_size[2] * config_size[3])
        u = xNSs.flatten()
        graph = GraphData(node_features=xSites, edge_index=np.array([u, v]))
        return graph


class _SiteEnvironment(object):

    def __init__(self,
                 pos: np.ndarray,
                 sitetypes: List[str],
                 env2config: List[int],
                 permutations: List[List[int]],
                 cutoff: float = 6.00,
                 Grtol: float = 0.0,
                 Gatol: float = 0.01,
                 rtol: float = 0.01,
                 atol: float = 0.0,
                 tol: float = 0.01,
                 grtol: float = 1e-3):
        """
        Initialize site environment

        This class contains local site environment information. This is used
        to find neighbor list in the datum.

        Parameters
        ----------
        pos : np.ndarray
            n x 3 list or numpy array of (non-scaled) positions. n is the
            number of atom.
        sitetypes : List[str]
            n list of string. String must be S or A followed by a
            number. S indicates a spectator sites and A indicates a active
            sites.
        env2config: List[int]
            A particular permutation of the neighbors around an active
            site. These indexes will be used for lattice transformation.
        permutations : List[List[int]]
            p x n list of list of integer. p is the permutation
            index and n is the number of sites.
        cutoff : float
            cutoff used for pooling neighbors.
        Grtol : float (default 0.0)
            relative tolerance in distance for forming an edge in graph
        Gatol : float (default 0.01)
            absolute tolerance in distance for forming an edge in graph
        rtol : float (default 0.01)
            relative tolerance in rmsd in distance for graph matching
        atol : float (default 0.0)
            absolute tolerance in rmsd in distance for graph matching
        tol : float (default 0.01)
            maximum tolerance of position RMSD to decide whether two
            environment are the same
        grtol : float (default 0.01)
            tolerance for deciding symmetric nodes
        """
        try:
            import networkx.algorithms.isomorphism as iso
        except:
            raise ImportError("This class requires networkx to be installed.")
        self.pos = pos
        self.sitetypes = sitetypes
        self.activesiteidx = [
            i for i, s in enumerate(self.sitetypes) if 'A' in s
        ]
        self.formula: DefaultDict[str, int] = defaultdict(int)
        for s in sitetypes:
            self.formula[s] += 1
        self.permutations = permutations
        self.env2config = env2config
        self.cutoff = cutoff
        # Set up site environment matcher
        self.tol = tol
        # Graphical option
        self.Grtol = Grtol
        self.Gatol = Gatol
        # tolerance for grouping nodes
        self.grtol = grtol
        # determine minimum distance between sitetypes.
        # This is used to determine the existence of an edge
        dists = squareform(pdist(pos))
        mindists = defaultdict(list)
        for i, row in enumerate(dists):
            row_dists = defaultdict(list)
            for j in range(0, len(sitetypes)):
                if i == j:
                    continue
                # Sort by bond
                row_dists[frozenset(
                    (sitetypes[i], sitetypes[j]))].append(dists[i, j])
            for pair in row_dists:
                mindists[pair].append(np.min(row_dists[pair]))
        # You want to maximize this in order to make sure every node gets an edge
        self.mindists = {}
        for pair in mindists:
            self.mindists[pair] = np.max(mindists[pair])
        # construct graph
        self.G = self._construct_graph(pos, sitetypes)
        # matcher options
        self._nm = iso.categorical_node_match('n', '')
        self._em = iso.numerical_edge_match('d', 0, rtol, 0)

    def _construct_graph(self, pos: np.ndarray, sitetypes: List[str]):
        """
        Returns local environment graph using networkx and
        tolerance specified.

        Parameters
        ----------
        pos: np.ndarray
            ns x 3. coordinates of positions. ns is the number of sites.
            sitetypes: ns. sitetype for each site
        sitetypes: List[str]
            List of site properties mentioning if it is a active site "Ai"
            or spectator site "Si".

        Returns
        ------
        G: networkx.classes.graph.Graph
            networkx graph used for matching site positions in
            datum.
        """
        try:
            import networkx as nx
        except:
            raise ImportError("This class requires networkx to be installed.")

        # construct graph
        G = nx.Graph()
        dists = cdist([[0, 0, 0]], pos - np.mean(pos, 0))[0]
        sdists = np.sort(dists)
        uniquedists = sdists[~(
            np.triu(np.abs(sdists[:, None] - sdists) <= self.grtol, 1)).any(0)]
        orderfromcenter = np.digitize(dists, uniquedists)
        # Add nodes
        for i, o in enumerate(orderfromcenter):
            G.add_node(i, n=str(o) + sitetypes[i])
        # Add edge. distance is edge attribute
        dists = pdist(pos)
        n = 0
        for i in range(len(sitetypes)):
            for j in range(i + 1, len(sitetypes)):
                if dists[n] < self.mindists[frozenset((sitetypes[i], sitetypes[j]))] or\
                    (abs(self.mindists[frozenset((sitetypes[i], sitetypes[j]))] - dists[n]) <= self.Gatol + self.Grtol * abs(dists[n])):
                    G.add_edge(i, j, d=dists[n])
                n += 1
        return G

    def get_mapping(self, env: Dict[str, Any]) -> Dict[int, int]:
        """
        Returns mapping of sites from input to this object

        Pymatgen molecule_matcher does not work unfortunately as it needs to be
        a reasonably physical molecule.
        Here, the graph is constructed by connecting the nearest neighbor, and
        isomorphism is performed to find matches, then kabsch algorithm is
        performed to make sure it is a match. NetworkX is used for portability.

        Parameters
        ----------

        env : Dict[str, Any]
            dictionary that contains information of local environment of a
            site in datum. See _get_SiteEnvironments definition in the class
            _SiteEnvironments for what this variable should be.

        Returns
        -------
        dict : Dict[int, int]
            Atom mapping from Primitive cell to data point.
        """
        try:
            import networkx.algorithms.isomorphism as iso
        except:
            raise ImportError("This class requires networkx to be installed.")
        # construct graph

        G = self._construct_graph(env['pos'], env['sitetypes'])
        if len(self.G.nodes) != len(G.nodes):
            s = 'Number of nodes is not equal.\n'
            raise ValueError(s)
        elif len(self.G.edges) != len(G.edges):
            logging.warning("Expected the number of edges to be equal",
                            len(self.G.edges), len(G.edges))
            s = 'Number of edges is not equal.\n'
            s += "- Is the data point's cell a redefined lattice of primitive cell?\n"
            s += '- If relaxed structure is used, you may want to check structure or increase Gatol\n'
            raise ValueError(s)
        GM = iso.GraphMatcher(self.G, G, self._nm, self._em)
        # Gets the isomorphic mapping. Also the most time consuming part of the code
        ams = list(GM.isomorphisms_iter())

        if not ams:
            s = 'No isomorphism found.\n'
            s += "- Is the data point's cell a redefined lattice of primitive cell?\n"
            s += '- If relaxed structure is used, you may want to check structure or increase rtol\n'
            raise ValueError(s)

        rmsd = []
        for am in ams:  # Loop over isomorphism
            # reconstruct graph after alinging point order
            xyz = np.zeros((len(self.pos), 3))
            for i in am:
                xyz[i, :] = env['pos'][am[i], :]
            rotation, _ = Rotation.align_vectors(self.pos, xyz)
            R = rotation.as_matrix()
            # RMSD
            rmsd.append(
                np.sqrt(
                    np.mean(
                        np.linalg.norm(np.dot(self.pos, R) - xyz, axis=1)**2)))
        mini = np.argmin(rmsd)
        minrmsd = rmsd[mini]
        if minrmsd < self.tol:
            return ams[mini]
        else:
            s = 'No isomorphism found.\n'
            s += '-Consider increasing neighbor finding tolerance'
            raise ValueError(s)


class _SiteEnvironments(object):

    def __init__(self, site_envs: List[_SiteEnvironment], ns: int, na: int,
                 aos: List[str], eigen_tol: float, pbc: np.typing.ArrayLike,
                 cutoff: float):
        """
        Initialize
        Use Load to initialize this class.

        Parameters
        ----------
        site_envs : List[_SiteEnvironment]
            list of _SiteEnvironment object
        ns : int
            number of spectator sites types
        na : int
            number of active sites types
        aos : List[str]
            Available occupational states for active sites
            string should be the name of the occupancy. (consistent with the input data)
        eigen_tol : float
            tolerance for eigenanalysis of point group analysis in pymatgen.
        pbc : ArrayLike
            Boolean array, periodic boundary condition.
        cutoff : float
            Cutoff radius in angstrom for pooling sites to construct local environment
        """
        self.site_envs = site_envs
        self.unique_site_types: List[str] = [
            env.sitetypes[0] for env in self.site_envs
        ]
        self.ns = ns
        self.na = na
        self.aos = aos
        self.eigen_tol = eigen_tol
        self.pbc = pbc
        self.cutoff = cutoff

    def read_datum(self,
                   struct,
                   cutoff_factor: float = 1.1) -> Tuple[np.ndarray, np.ndarray]:
        """
        Load structure data and return neighbor information

        Parameters
        ----------
        struct: : PymatgenStructure
            Pymatgen Structure object of the surface configuration. It also requires
            site_properties attribute with "Sitetypes"(Active or spectator site) and
            "oss"(Species of Active site from the list of self.aos and "-1" for
            spectator sites).
        cutoff_factor : float
            this is extra buffer factor multiplied to cutoff to
            ensure pooling all relevant sites.

        Return
        ------
        XSites : List[float]
            One hot encoding features of the site.
        XNSs : List[List[int]]
            Neighbors calculated in different permutations
        """
        oss = [
            species for species in struct.site_properties["oss"]
            if species != '-1'
        ]
        # Construct one hot encoding
        XSites = np.zeros((len(oss), len(self.aos)))
        for i, o in enumerate(oss):
            XSites[i, self.aos.index(o)] = 1
        # get mapping between all site index to active site index
        alltoactive = {}
        n = 0
        for i, s in enumerate(struct.site_properties["SiteTypes"]):
            if 'A' in s:
                alltoactive[i] = n
                n += 1
        # Get Neighbors
        # Read Data
        site_envs = _get_SiteEnvironments(struct,
                                          self.cutoff * cutoff_factor,
                                          self.pbc,
                                          get_permutations=False,
                                          eigen_tol=self.eigen_tol)
        XNSs: List[list] = [[] for _ in range(len(self.site_envs))]
        for env in site_envs:
            i = self.unique_site_types.index(env['sitetypes'][0])
            new_env = self._truncate(self.site_envs[i], env)

            # get map between two environment
            mapping = self.site_envs[i].get_mapping(new_env)
            # align input to the primitive cell (reference)
            aligned_idx = [
                new_env['env2config'][mapping[i]]
                for i in range(len(new_env['env2config']))
            ]
            # apply permutations
            nni_perm = np.take(aligned_idx, self.site_envs[i].permutations)
            # remove spectators
            nni_perm = nni_perm[:, self.site_envs[i].activesiteidx]
            # map it to active sites
            nni_perm = np.vectorize(alltoactive.__getitem__)(nni_perm)
            XNSs[i].append(nni_perm.tolist())
        return np.array(XSites), np.array(XNSs)

    @classmethod
    def _truncate(cls, env_ref: _SiteEnvironment,
                  env: Dict[str, Any]) -> Dict[str, Any]:
        """
        When cutoff_factor is used, it will pool more site than cutoff
        factor specifies. This will rule out non-relevant sites by distance.

        Parameters
        ----------
        env_ref: _SiteEnvironment
            Site information of the primitive cell
        env: Dict[str, Any]
            Site information of the data point

        Returns
        -------
        env: Dict[str, Union[list, np.ndarray]]
        """
        # Extract the right number of sites by distance
        dists = defaultdict(list)
        for i, s in enumerate(env['sitetypes']):
            dists[s].append([i, env['dist'][i]])
        for s in dists:
            dists[s] = sorted(dists[s], key=lambda x: x[1])
        siteidx = []
        for s in dists:
            siteidx += [i[0] for i in dists[s][:env_ref.formula[s]]]
        siteidx = sorted(siteidx)
        env['pos'] = [
            env['pos'][i] for i in range(len(env['pos'])) if i in siteidx
        ]

        env['pos'] = np.subtract(env['pos'], np.mean(env['pos'], 0))
        env['sitetypes'] = [
            env['sitetypes'][i]
            for i in range(len(env['sitetypes']))
            if i in siteidx
        ]
        env['env2config'] = [env['env2config'][i] for i in siteidx]
        del env['dist']
        return env


def _load_primitive_cell(struct: PymatgenStructure,
                         aos: List[str],
                         pbc: List[bool],
                         ns: int,
                         na: int,
                         cutoff: float,
                         eigen_tol: float = 1e-5) -> _SiteEnvironments:
    """
    This loads the primitive cell, along with all the permutations
    required for creating a neighbor. This produces the site environments of
    the primitive cell.

    Parameters
    ----------
    struct: PymatgenStructure
        Pymatgen Structure object of the primitive cell used for calculating
        neighbors from lattice transformations.It also requires site_properties
        attribute with "Sitetypes"(Active or spectator site).
    aos: List[str]
        A list of all the active site species. For the Pt, N, NO configuration
        set it as ['0', '1', '2'].
    pbc: List[bool]
        Periodic Boundary Condition
    ns: int (default 1)
        The number of spectator types elements. For "S1" its 1.
    na: int (default 1)
        The number of active types elements. For "A1" its 1.
    cutoff: float (default 6.00)
        Cutoff of radius for getting local environment.Only
        used down to 2 digits.
    eigen_tol : float (default)
        tolerance for eigenanalysis of point group analysis in
        pymatgen.

    Returns
    -------
    SiteEnvironments: _SiteEnvironments
        Instance of the _SiteEnvironments object
    """
    site_envs = _get_SiteEnvironments(struct,
                                      cutoff,
                                      pbc,
                                      True,
                                      eigen_tol=eigen_tol)
    site_envs_format = [
        _SiteEnvironment(e['pos'], e['sitetypes'], e['env2config'],
                         e['permutations'], cutoff) for e in site_envs
    ]

    ust = [env.sitetypes[0] for env in site_envs_format]
    usi = np.unique(ust, return_index=True)[1]
    site_envs_format = [site_envs_format[i] for i in usi]
    return _SiteEnvironments(site_envs_format, ns, na, aos, eigen_tol, pbc,
                             cutoff)


def _get_SiteEnvironments(struct: PymatgenStructure,
                          cutoff: float,
                          PBC: np.typing.ArrayLike,
                          get_permutations: bool = True,
                          eigen_tol: float = 1e-5) -> List[Dict[str, Any]]:
    """
    Used to extract information about both primitive cells and data points.
    Extract local environments from Structure object by calculating neighbors
    based on gaussian distance. For primitive cell, Different permutations of the
    neighbors are calculated and will be later will mapped for data point in the
    _SiteEnvironment.get_mapping() function.
    site types ,

    Parameters
    ----------
    struct: PymatgenStructure
        Pymatgen Structure object of the primitive cell used for calculating
        neighbors from lattice transformations.It also requires site_properties
        attribute with "Sitetypes"(Active or spectator site).
    cutoff : float
        cutoff distance in angstrom for collecting local
        environment.
    pbc : ArrayLike
        Periodic boundary condition
    get_permutations : bool (default True)
        Whether to find permuted neighbor list or not.
    eigen_tol : float (default 1e-5)
        Tolerance for eigenanalysis of point group analysis in
        pymatgen.

    Returns
    ------
    site_envs : List[Dict[str, Any]]
        list of local_env class
    """
    try:
        from pymatgen.core import Molecule
        from pymatgen.symmetry.analyzer import PointGroupAnalyzer
    except:
        raise ImportError("This class requires pymatgen to be installed.")

    pbc = np.array(PBC)
    structure = struct
    neighbors = structure.get_all_neighbors(cutoff, include_index=True)
    symbols = structure.species
    site_idxs = [
        i for i, sitetype in enumerate(structure.site_properties['SiteTypes'])
        if sitetype == 'A1'
    ]
    site_sym_map = {}
    sym_site_map = {}
    for i, new_ele in enumerate(structure.species):
        sym_site_map[new_ele] = structure.site_properties['SiteTypes'][i]
        site_sym_map[structure.site_properties['SiteTypes'][i]] = new_ele

    site_envs = []
    for site_idx in site_idxs:
        local_env_sym = [symbols[site_idx]]
        local_env_xyz = [structure[site_idx].coords]
        local_env_dist = [0.0]
        local_env_sitemap = [site_idx]
        for n in neighbors[site_idx]:
            # if PBC condition is fulfilled..
            c = np.around(n[0].frac_coords, 10)
            withinPBC = np.logical_and(0 <= c, c < 1)
            if np.all(withinPBC[~pbc]):
                local_env_xyz.append(n[0].coords)
                local_env_sym.append(n[0].specie)
                local_env_dist.append(n[1])
                local_env_sitemap.append(n[2])
        local_env_pos = np.subtract(local_env_xyz, np.mean(local_env_xyz, 0))

        perm = []
        if get_permutations:
            finder = PointGroupAnalyzer(
                Molecule(local_env_sym, local_env_pos),  # type: ignore
                eigen_tolerance=eigen_tol)
            pg = finder.get_pointgroup()
            for i, op in enumerate(pg):
                newpos = op.operate_multi(local_env_pos)
                perm.append(
                    np.argmin(cdist(local_env_pos, newpos), axis=1).tolist())

        site_env = {
            'pos': local_env_pos,
            'sitetypes': [sym_site_map[s] for s in local_env_sym],
            'env2config': local_env_sitemap,
            'permutations': perm,
            'dist': local_env_dist
        }
        site_envs.append(site_env)
    return site_envs



# flake8:noqa
import logging

logger = logging.getLogger(__name__)

from deepchem.models.torch_models.torch_model import TorchModel
from deepchem.models.torch_models.modular import ModularTorchModel
from deepchem.models.torch_models.attentivefp import AttentiveFP, AttentiveFPModel
from deepchem.models.torch_models.cgcnn import CGCNN, CGCNNModel
from deepchem.models.torch_models.gat import GAT, GATModel
from deepchem.models.torch_models.gcn import GCN, GCNModel
from deepchem.models.torch_models.infograph import InfoGraphStar, InfoGraphStarModel, InfoGraphEncoder, GINEncoder, InfoGraph, InfoGraphModel, InfoGraphEncoder
from deepchem.models.torch_models.mpnn import MPNN, MPNNModel
from deepchem.models.torch_models.lcnn import LCNN, LCNNModel
from deepchem.models.torch_models.pagtn import Pagtn, PagtnModel
from deepchem.models.torch_models.mat import MAT, MATModel
from deepchem.models.torch_models.megnet import MEGNetModel
from deepchem.models.torch_models.normalizing_flows_pytorch import NormalizingFlow
from deepchem.models.torch_models.layers import MultilayerPerceptron, CNNModule, CombineMeanStd, WeightedLinearCombo, AtomicConvolution, NeighborList, SetGather
from deepchem.models.torch_models.cnn import CNN
from deepchem.models.torch_models.attention import ScaledDotProductAttention, SelfAttention
from deepchem.models.torch_models.grover import GroverModel, GroverPretrain, GroverFinetune
from deepchem.models.torch_models.readout import GroverReadout
try:
    from deepchem.models.torch_models.dmpnn import DMPNN, DMPNNModel
    from deepchem.models.torch_models.gnn import GNN, GNNHead, GNNModular
    from deepchem.models.torch_models.pna_gnn import AtomEncoder, BondEncoder, PNALayer, PNAGNN, PNA
    from deepchem.models.torch_models.gnn3d import Net3D, InfoMax3DModular
except ModuleNotFoundError as e:
    logger.warning(
        f'Skipped loading modules with pytorch-geometric dependency, missing a dependency. {e}'
    )
try:
    from deepchem.models.torch_models.hf_models import HuggingFaceModel
    from deepchem.models.torch_models.chemberta import Chemberta
except ModuleNotFoundError as e:
    logger.warning(f'Skipped loading modules with transformers dependency. {e}')



"""
Test Autodock Vina Utility Functions.
"""
import os
import numpy as np
import unittest
from deepchem.utils import docking_utils
from deepchem.utils import rdkit_utils


class TestVinaUtils(unittest.TestCase):

    def setUp(self):
        # TODO test more formats for ligand
        self.current_dir = os.path.dirname(os.path.realpath(__file__))
        self.docked_ligands = os.path.join(self.current_dir, 'assets',
                                           '1jld_ligand_docked.pdbqt')

    def test_load_docked_ligand(self):
        docked_ligands, scores = docking_utils.load_docked_ligands(
            self.docked_ligands)
        assert len(docked_ligands) == 9
        assert len(scores) == 9

        for ligand, score in zip(docked_ligands, scores):
            xyz = rdkit_utils.get_xyz_from_mol(ligand)
            assert score < 0  # This is a binding free energy
            assert np.count_nonzero(xyz) > 0

    def test_write_gnina_conf(self):
        docking_utils.write_gnina_conf(
            'protein.pdb',
            'ligand.sdf',
            'conf.txt',
        )
        assert os.path.exists('conf.txt')
        os.remove('conf.txt')

    def test_read_gnina_log(self):
        log_file = os.path.join(self.current_dir, 'assets', 'gnina_log.txt')
        scores = docking_utils.read_gnina_log(log_file)
        assert np.array_equal(
            scores, np.array([[-4.37, 0.6392, 4.336], [-3.56, 0.6202, 4.162]]))

    def test_prepare_inputs(self):
        pdbid = '3cyx'
        ligand_smiles = 'CC(C)(C)NC(O)C1CC2CCCCC2C[NH+]1CC(O)C(CC1CCCCC1)NC(O)C(CC(N)O)NC(O)C1CCC2CCCCC2N1'

        protein, ligand = docking_utils.prepare_inputs(pdbid,
                                                       ligand_smiles,
                                                       pdb_name=pdbid)

        assert np.isclose(protein.GetNumAtoms(), 1510, atol=3)
        assert np.isclose(ligand.GetNumAtoms(), 124, atol=3)

        protein, ligand = docking_utils.prepare_inputs(
            pdbid + '.pdb', 'ligand_' + pdbid + '.pdb')

        assert np.isclose(protein.GetNumAtoms(), 1510, atol=3)
        assert np.isclose(ligand.GetNumAtoms(), 124, atol=3)

        os.remove(pdbid + '.pdb')
        os.remove('ligand_' + pdbid + '.pdb')



"""
Atomic coordinate featurizer.
"""
import logging
import warnings

import numpy as np

from deepchem.feat.base_classes import Featurizer, ComplexFeaturizer
from deepchem.feat.molecule_featurizers import AtomicCoordinates
from deepchem.utils.data_utils import pad_array
from deepchem.utils.rdkit_utils import MoleculeLoadException, get_xyz_from_mol, \
    load_molecule, merge_molecules_xyz, merge_molecules


def compute_neighbor_list(coords, neighbor_cutoff, max_num_neighbors,
                          periodic_box_size):
    """Computes a neighbor list from atom coordinates."""
    N = coords.shape[0]
    try:
        import mdtraj
    except ModuleNotFoundError:
        raise ImportError("This function requires mdtraj to be installed.")
    traj = mdtraj.Trajectory(coords.reshape((1, N, 3)), None)
    box_size = None
    if periodic_box_size is not None:
        box_size = np.array(periodic_box_size)
        traj.unitcell_vectors = np.array(
            [[[box_size[0], 0, 0], [0, box_size[1], 0], [0, 0, box_size[2]]]],
            dtype=np.float32)
    neighbors = mdtraj.geometry.compute_neighborlist(traj, neighbor_cutoff)
    neighbor_list = {}
    for i in range(N):
        if max_num_neighbors is not None and len(
                neighbors[i]) > max_num_neighbors:
            delta = coords[i] - coords.take(neighbors[i], axis=0)
            if box_size is not None:
                delta -= np.round(delta / box_size) * box_size
            dist = np.linalg.norm(delta, axis=1)
            sorted_neighbors = list(zip(dist, neighbors[i]))
            sorted_neighbors.sort()
            neighbor_list[i] = [
                sorted_neighbors[j][1] for j in range(max_num_neighbors)
            ]
        else:
            neighbor_list[i] = list(neighbors[i])
    return neighbor_list


class NeighborListAtomicCoordinates(Featurizer):
    """
    Adjacency List of neighbors in 3-space

    Neighbors determined by user-defined distance cutoff [in Angstrom].

    https://en.wikipedia.org/wiki/Cell_list
    Ref: http://www.cs.cornell.edu/ron/references/1989/Calculations%20of%20a%20List%20of%20Neighbors%20in%20Molecular%20Dynamics%20Si.pdf

    Parameters
    ----------
    neighbor_cutoff: float
        Threshold distance [Angstroms] for counting neighbors.
    periodic_box_size: 3 element array
        Dimensions of the periodic box in Angstroms, or None to not use periodic boundary conditions
    """

    def __init__(self,
                 max_num_neighbors=None,
                 neighbor_cutoff=4,
                 periodic_box_size=None):
        if neighbor_cutoff <= 0:
            raise ValueError("neighbor_cutoff must be positive value.")
        if max_num_neighbors is not None:
            if not isinstance(max_num_neighbors, int) or max_num_neighbors <= 0:
                raise ValueError("max_num_neighbors must be positive integer.")
        self.max_num_neighbors = max_num_neighbors
        self.neighbor_cutoff = neighbor_cutoff
        self.periodic_box_size = periodic_box_size
        # Type of data created by this featurizer
        self.dtype = object
        self.bohr_coords_featurizer = AtomicCoordinates(use_bohr=True)
        self.coords_featurizer = AtomicCoordinates(use_bohr=False)

    def _featurize(self, mol):
        """
        Compute neighbor list.

        Parameters
        ----------
        mol: rdkit Mol
            To be featurized.
        """
        # TODO(rbharath): Should this return a list?
        bohr_coords = self.bohr_coords_featurizer._featurize(mol)
        coords = self.coords_featurizer._featurize(mol)
        neighbor_list = compute_neighbor_list(coords, self.neighbor_cutoff,
                                              self.max_num_neighbors,
                                              self.periodic_box_size)
        return (bohr_coords, neighbor_list)


class NeighborListComplexAtomicCoordinates(ComplexFeaturizer):
    """
    Adjacency list of neighbors for protein-ligand complexes in 3-space.

    Neighbors determined by user-defined distance cutoff.
    """

    def __init__(self, max_num_neighbors=None, neighbor_cutoff=4):
        if neighbor_cutoff <= 0:
            raise ValueError("neighbor_cutoff must be positive value.")
        if max_num_neighbors is not None:
            if not isinstance(max_num_neighbors, int) or max_num_neighbors <= 0:
                raise ValueError("max_num_neighbors must be positive integer.")
        self.max_num_neighbors = max_num_neighbors
        self.neighbor_cutoff = neighbor_cutoff
        # Type of data created by this featurizer
        self.dtype = object

    def _featurize(self, datapoint, **kwargs):
        """
        Compute neighbor list for complex.

        Parameters
        ----------
        datapoint: Tuple[str, str]
            Filenames for molecule and protein.
        """
        if 'complex' in kwargs:
            datapoint = kwargs.get("complex")
            raise DeprecationWarning(
                'Complex is being phased out as a parameter, please pass "datapoint" instead.'
            )

        mol_pdb_file, protein_pdb_file = datapoint
        mol_coords, ob_mol = load_molecule(mol_pdb_file)
        protein_coords, protein_mol = load_molecule(protein_pdb_file)
        system_coords = merge_molecules_xyz([mol_coords, protein_coords])

        system_neighbor_list = compute_neighbor_list(system_coords,
                                                     self.neighbor_cutoff,
                                                     self.max_num_neighbors,
                                                     None)

        return (system_coords, system_neighbor_list)


class AtomicConvFeaturizer(ComplexFeaturizer):
    """This class computes the featurization that corresponds to AtomicConvModel.

    This class computes featurizations needed for AtomicConvModel.
    Given two molecular structures, it computes a number of useful
    geometric features. In particular, for each molecule and the global
    complex, it computes a coordinates matrix of size (N_atoms, 3)
    where N_atoms is the number of atoms. It also computes a
    neighbor-list, a dictionary with N_atoms elements where
    neighbor-list[i] is a list of the atoms the i-th atom has as
    neighbors. In addition, it computes a z-matrix for the molecule
    which is an array of shape (N_atoms,) that contains the atomic
    number of that atom.

    Since the featurization computes these three quantities for each of
    the two molecules and the complex, a total of 9 quantities are
    returned for each complex. Note that for efficiency, fragments of
    the molecules can be provided rather than the full molecules
    themselves.

    """

    def __init__(self,
                 frag1_num_atoms,
                 frag2_num_atoms,
                 complex_num_atoms,
                 max_num_neighbors,
                 neighbor_cutoff,
                 strip_hydrogens=True):
        """

        Parameters
        ----------
        frag1_num_atoms: int
            Maximum number of atoms in fragment 1.
        frag2_num_atoms: int
            Maximum number of atoms in fragment 2.
        complex_num_atoms: int
            Maximum number of atoms in complex of frag1/frag2 together.
        max_num_neighbors: int
            Maximum number of atoms considered as neighbors.
        neighbor_cutoff: float
            Maximum distance (angstroms) for two atoms to be considered as
            neighbors. If more than `max_num_neighbors` atoms fall within
            this cutoff, the closest `max_num_neighbors` will be used.
        strip_hydrogens: bool (default True)
            Remove hydrogens before computing featurization.

        """

        self.frag1_num_atoms = frag1_num_atoms
        self.frag2_num_atoms = frag2_num_atoms
        self.complex_num_atoms = complex_num_atoms
        self.max_num_neighbors = max_num_neighbors
        self.neighbor_cutoff = neighbor_cutoff
        self.strip_hydrogens = strip_hydrogens
        self.neighborlist_featurizer = NeighborListComplexAtomicCoordinates(
            self.max_num_neighbors, self.neighbor_cutoff)

    def _featurize(self, complex):
        mol_pdb_file, protein_pdb_file = complex
        try:
            frag1_coords, frag1_mol = load_molecule(mol_pdb_file,
                                                    is_protein=False,
                                                    sanitize=True,
                                                    add_hydrogens=False)
            frag2_coords, frag2_mol = load_molecule(protein_pdb_file,
                                                    is_protein=True,
                                                    sanitize=True,
                                                    add_hydrogens=False)

        except MoleculeLoadException:
            # Currently handles loading failures by returning None
            # TODO: Is there a better handling procedure?
            logging.warning(
                "Some molecules cannot be loaded by Rdkit. Skipping")
            return None

        except ImportError as e:
            logging.warning("%s" % e)
            raise ImportError(e)

        system_mol = merge_molecules([frag1_mol, frag2_mol])
        system_coords = get_xyz_from_mol(system_mol)

        frag1_coords, frag1_mol = self._strip_hydrogens(frag1_coords, frag1_mol)
        frag2_coords, frag2_mol = self._strip_hydrogens(frag2_coords, frag2_mol)
        system_coords, system_mol = self._strip_hydrogens(
            system_coords, system_mol)

        try:
            frag1_coords, frag1_neighbor_list, frag1_z = self.featurize_mol(
                frag1_coords, frag1_mol, self.frag1_num_atoms)

            frag2_coords, frag2_neighbor_list, frag2_z = self.featurize_mol(
                frag2_coords, frag2_mol, self.frag2_num_atoms)

            system_coords, system_neighbor_list, system_z = self.featurize_mol(
                system_coords, system_mol, self.complex_num_atoms)
        except ValueError:
            logging.warning(
                "max_atoms was set too low. Some complexes too large and skipped"
            )
            return None
        except ImportError as e:
            logging.warning("%s" % e)
            raise ImportError(e)

        return frag1_coords, frag1_neighbor_list, frag1_z, frag2_coords, frag2_neighbor_list, frag2_z, \
               system_coords, system_neighbor_list, system_z

    def get_Z_matrix(self, mol, max_atoms):
        if len(mol.GetAtoms()) > max_atoms:
            raise ValueError(
                "A molecule is larger than permitted by max_atoms. "
                "Increase max_atoms and try again.")
        return pad_array(
            np.array([atom.GetAtomicNum() for atom in mol.GetAtoms()]),
            max_atoms)

    def featurize_mol(self, coords, mol, max_num_atoms):
        logging.info("Featurizing molecule of size: %d", len(mol.GetAtoms()))
        neighbor_list = compute_neighbor_list(coords, self.neighbor_cutoff,
                                              self.max_num_neighbors, None)
        # pad outputs
        z = self.get_Z_matrix(mol, max_num_atoms)
        z = pad_array(z, max_num_atoms)
        coords = pad_array(coords, (max_num_atoms, 3))
        return coords, neighbor_list, z

    def _strip_hydrogens(self, coords, mol):

        class MoleculeShim(object):
            """
            Shim of a Molecule which supports #GetAtoms()
            """

            def __init__(self, atoms):
                self.atoms = [AtomShim(x) for x in atoms]

            def GetAtoms(self):
                return self.atoms

        class AtomShim(object):

            def __init__(self, atomic_num):
                self.atomic_num = atomic_num

            def GetAtomicNum(self):
                return self.atomic_num

        if not self.strip_hydrogens:
            return coords, mol
        indexes_to_keep = []
        atomic_numbers = []
        for index, atom in enumerate(mol.GetAtoms()):
            if atom.GetAtomicNum() != 1:
                indexes_to_keep.append(index)
                atomic_numbers.append(atom.GetAtomicNum())
        mol = MoleculeShim(atomic_numbers)
        coords = coords[indexes_to_keep]
        return coords, mol


# Deprecation warnings for old atomic conv featurizer name #

ATOMICCONV_DEPRECATION = "{} is deprecated and has been renamed to {} and will be removed in DeepChem 3.0."


class ComplexNeighborListFragmentAtomicCoordinates(AtomicConvFeaturizer):

    def __init__(self, *args, **kwargs):

        warnings.warn(
            ATOMICCONV_DEPRECATION.format(
                "ComplexNeighorListFragmentAtomicCoordinates",
                "AtomicConvFeaturizer"), FutureWarning)

        super(ComplexNeighborListFragmentAtomicCoordinates,
              self).__init__(*args, **kwargs)



import copy
from functools import partial
from math import sqrt
from typing import Callable, Dict, List, Union

import dgl
import torch
from torch import nn

from deepchem.feat.molecule_featurizers.conformer_featurizer import (
    full_atom_feature_dims,
    full_bond_feature_dims,
)
from deepchem.models.torch_models.layers import MultilayerPerceptron
from deepchem.utils.graph_utils import (
    aggregate_max,
    aggregate_mean,
    aggregate_min,
    aggregate_moment,
    aggregate_std,
    aggregate_sum,
    aggregate_var,
    scale_amplification,
    scale_attenuation,
    scale_identity,
)

PNA_AGGREGATORS = {
    "mean": aggregate_mean,
    "sum": aggregate_sum,
    "max": aggregate_max,
    "min": aggregate_min,
    "std": aggregate_std,
    "var": aggregate_var,
    "moment3": partial(aggregate_moment, n=3),
    "moment4": partial(aggregate_moment, n=4),
    "moment5": partial(aggregate_moment, n=5),
}

PNA_SCALERS = {
    "identity": scale_identity,
    "amplification": scale_amplification,
    "attenuation": scale_attenuation,
}


class AtomEncoder(torch.nn.Module):
    """
    Encodes atom features into embeddings based on the Open Graph Benchmark feature set in conformer_featurizer.

    Parameters
    ----------
    emb_dim : int
        The dimension that the returned embedding will have.
    padding : bool, optional (default=False)
        If true then the last index will be used for padding.

    Examples
    --------
    >>> from deepchem.feat.molecule_featurizers.conformer_featurizer import full_atom_feature_dims
    >>> atom_encoder = AtomEncoder(emb_dim=32)
    >>> num_rows = 10
    >>> atom_features = torch.stack([
    ... torch.randint(low=0, high=dim, size=(num_rows,))
    ... for dim in full_atom_feature_dims
    ... ], dim=1)
    >>> atom_embeddings = atom_encoder(atom_features)
    """

    def __init__(self, emb_dim, padding=False):
        super(AtomEncoder, self).__init__()

        self.atom_embedding_list = torch.nn.ModuleList()
        self.padding = padding

        for dim in full_atom_feature_dims:
            if padding:
                emb = torch.nn.Embedding(dim + 1, emb_dim, padding_idx=0)
            else:
                emb = torch.nn.Embedding(dim, emb_dim)
            torch.nn.init.xavier_uniform_(emb.weight.data)
            self.atom_embedding_list.append(emb)

    def reset_parameters(self):
        """
        Reset the parameters of the atom embeddings.

        This method resets the weights of the atom embeddings by initializing
        them with a uniform distribution between -sqrt(3) and sqrt(3).
        """
        for embedder in self.atom_embedding_list:
            embedder.weight.data.uniform_(-sqrt(3), sqrt(3))

    def forward(self, x):
        """
        Compute the atom embeddings for the given atom features.

        Parameters
        ----------
        x : torch.Tensor, shape (batch_size, num_atoms, num_features)
            The input atom features tensor.

        Returns
        -------
        x_embedding : torch.Tensor, shape (batch_size, num_atoms, emb_dim)
            The computed atom embeddings.
        """
        x_embedding = 0
        for i in range(x.shape[1]):
            if self.padding:
                x_embedding += self.atom_embedding_list[i](x[:, i].long() + 1)
            else:
                x_embedding += self.atom_embedding_list[i](x[:, i].long())

        return x_embedding


class BondEncoder(torch.nn.Module):
    """
    Encodes bond features into embeddings based on the Open Graph Benchmark feature set in conformer_featurizer.

    Parameters
    ----------
    emb_dim : int
        The dimension that the returned embedding will have.
    padding : bool, optional (default=False)
        If true then the last index will be used for padding.

    Examples
    --------
    >>> from deepchem.feat.molecule_featurizers.conformer_featurizer import full_bond_feature_dims
    >>> bond_encoder = BondEncoder(emb_dim=32)
    >>> num_rows = 10
    >>> bond_features = torch.stack([
    ... torch.randint(low=0, high=dim, size=(num_rows,))
    ... for dim in full_bond_feature_dims
    ... ], dim=1)
    >>> bond_embeddings = bond_encoder(bond_features)
    """

    def __init__(self, emb_dim, padding=False):
        super(BondEncoder, self).__init__()

        self.bond_embedding_list = torch.nn.ModuleList()
        self.padding = padding

        for dim in full_bond_feature_dims:
            if padding:
                emb = torch.nn.Embedding(dim + 1, emb_dim, padding_idx=0)
            else:
                emb = torch.nn.Embedding(dim, emb_dim)
            torch.nn.init.xavier_uniform_(emb.weight.data)
            self.bond_embedding_list.append(emb)

    def forward(self, edge_attr):
        """
        Compute the bond embeddings for the given bond features.

        Parameters
        ----------
        edge_attr : torch.Tensor, shape (batch_size, num_edges, num_features)
            The input bond features tensor.

        Returns
        -------
        bond_embedding : torch.Tensor, shape (batch_size, num_edges, emb_dim)
            The computed bond embeddings.
        """
        bond_embedding = 0
        for i in range(edge_attr.shape[1]):
            if self.padding:
                bond_embedding += self.bond_embedding_list[i](
                    edge_attr[:, i].long() + 1)
            else:
                bond_embedding += self.bond_embedding_list[i](
                    edge_attr[:, i].long())

        return bond_embedding


class PNALayer(nn.Module):
    """
    Principal Neighbourhood Aggregation Layer (PNA) from [1].

    Parameters
    ----------
    in_dim : int
        Input dimension of the node features.
    out_dim : int
        Output dimension of the node features.
    in_dim_edges : int
        Input dimension of the edge features.
    aggregators : List[str]
        List of aggregator functions to use. Options are "mean", "sum", "max", "min", "std", "var", "moment3", "moment4", "moment5".
    scalers : List[str]
        List of scaler functions to use. Options are "identity", "amplification", "attenuation".
    activation : Union[Callable, str], optional, default="relu"
        Activation function to use.
    last_activation : Union[Callable, str], optional, default="none"
        Last activation function to use.
    dropout : float, optional, default=0.0
        Dropout rate.
    residual : bool, optional, default=True
        Whether to use residual connections.
    pairwise_distances : bool, optional, default=False
        Whether to use pairwise distances.
    batch_norm : bool, optional, default=True
        Whether to use batch normalization.
    batch_norm_momentum : float, optional, default=0.1
        Momentum for the batch normalization layers.
    avg_d : Dict[str, float], optional, default={"log": 1.0}
        Dictionary containing the average degree of the graph.
    posttrans_layers : int, optional, default=2
        Number of post-transformation layers.
    pretrans_layers : int, optional, default=1
        Number of pre-transformation layers.

    References
    ----------
    .. [1] Corso, G., Cavalleri, L., Beaini, D., Liò, P. & Veličković, P. Principal Neighbourhood Aggregation for Graph Nets. Preprint at https://doi.org/10.48550/arXiv.2004.05718 (2020).

    Examples
    --------
    >>> import dgl
    >>> import numpy as np
    >>> import torch
    >>> from deepchem.models.torch_models.pna_gnn import PNALayer
    >>> in_dim = 32
    >>> out_dim = 64
    >>> in_dim_edges = 16
    >>> aggregators = ["mean", "max"]
    >>> scalers = ["identity", "amplification", "attenuation"]
    >>> pna_layer = PNALayer(in_dim=in_dim,
    ...                      out_dim=out_dim,
    ...                      in_dim_edges=in_dim_edges,
    ...                      aggregators=aggregators,
    ...                      scalers=scalers)
    >>> num_nodes = 10
    >>> num_edges = 20
    >>> node_features = torch.randn(num_nodes, in_dim)
    >>> edge_features = torch.randn(num_edges, in_dim_edges)
    >>> g = dgl.graph((np.random.randint(0, num_nodes, num_edges),
    ...                np.random.randint(0, num_nodes, num_edges)))
    >>> g.ndata['feat'] = node_features
    >>> g.edata['feat'] = edge_features
    >>> g.ndata['feat'] = pna_layer(g)
    """

    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        in_dim_edges: int,
        aggregators: List[str],
        scalers: List[str],
        activation: Union[Callable, str] = "relu",
        dropout: float = 0.0,
        residual: bool = True,
        pairwise_distances: bool = False,
        batch_norm: bool = True,
        batch_norm_momentum=0.1,
        avg_d: Dict[str, float] = {"log": 1.0},
        posttrans_layers: int = 2,
        pretrans_layers: int = 1,
    ):
        super(PNALayer, self).__init__()
        self.aggregators = [PNA_AGGREGATORS[aggr] for aggr in aggregators]
        self.scalers = [PNA_SCALERS[scale] for scale in scalers]
        self.edge_features = in_dim_edges > 0
        self.activation = activation
        self.avg_d = avg_d
        self.pairwise_distances = pairwise_distances
        self.residual = residual
        if in_dim != out_dim:
            self.residual = False

        self.pretrans = MultilayerPerceptron(
            d_input=(2 * in_dim + in_dim_edges +
                     1) if self.pairwise_distances else
            (2 * in_dim + in_dim_edges),
            d_output=in_dim,
            d_hidden=(in_dim,) * (pretrans_layers - 1),
            batch_norm=batch_norm,
            batch_norm_momentum=batch_norm_momentum,
            dropout=dropout)

        self.posttrans = MultilayerPerceptron(
            d_input=(len(self.aggregators) * len(self.scalers) + 1) * in_dim,
            d_hidden=(out_dim,) * (posttrans_layers - 1),
            d_output=out_dim,
            batch_norm=batch_norm,
            batch_norm_momentum=batch_norm_momentum,
            dropout=dropout)

    def forward(self, g):
        """
        Forward pass of the PNA layer.

        Parameters
        ----------
        g : dgl.DGLGraph
            Input graph

        Returns
        -------
        h : torch.Tensor
            Node feature tensor
        """
        h = g.ndata['feat']
        h_in = h
        # pretransformation
        g.apply_edges(self.pretrans_edges)

        # aggregation
        g.update_all(self.message_func, self.reduce_func)
        h = torch.cat([h, g.ndata['feat']], dim=-1)
        # post-transformation
        h = self.posttrans(h)
        if self.residual:
            h = h + h_in

        return h

    def message_func(self, edges) -> Dict[str, torch.Tensor]:
        """
        The message function to generate messages along the edges.

        Parameters
        ----------
        edges : dgl.EdgeBatch
            Batch of edges.

        Returns
        -------
        Dict[str, torch.Tensor]
            Dictionary containing the edge features.
        """
        return {"e": edges.data["e"]}

    def reduce_func(self, nodes) -> Dict[str, torch.Tensor]:
        """
        The reduce function to aggregate the messages.
        Apply the aggregators and scalers, and concatenate the results.

        Parameters
        ----------
        nodes : dgl.NodeBatch
            Batch of nodes.

        Returns
        -------
        Dict[str, torch.Tensor]
            Dictionary containing the aggregated node features.
        """
        h_in = nodes.data['feat']
        h = nodes.mailbox["e"]
        D = h.shape[-2]

        h_to_cat = [
            aggr(h=h, h_in=h_in)  # type: ignore
            for aggr in self.aggregators
        ]
        h = torch.cat(h_to_cat, dim=-1)

        if len(self.scalers) > 1:
            h = torch.cat(
                [
                    scale(h, D=D, avg_d=self.avg_d)  # type: ignore
                    for scale in self.scalers
                ],
                dim=-1)

        return {'feat': h}

    def pretrans_edges(self, edges) -> Dict[str, torch.Tensor]:
        """
        Return a mapping to the concatenation of the features from
        the source node, the destination node, and the edge between them (if applicable).

        Parameters
        ----------
        edges : dgl.EdgeBatch
            Batch of edges.

        Returns
        -------
        Dict[str, torch.Tensor]
            Dictionary containing the concatenated features.
        """

        if self.edge_features and self.pairwise_distances:
            squared_distance = torch.sum((edges.src['x'] - edges.dst['x'])**2,
                                         dim=-1)[:, None]
            z2 = torch.cat([
                edges.src['feat'], edges.dst['feat'], edges.data['feat'],
                squared_distance
            ],
                           dim=-1)
        elif not self.edge_features and self.pairwise_distances:
            squared_distance = torch.sum((edges.src['x'] - edges.dst['x'])**2,
                                         dim=-1)[:, None]
            z2 = torch.cat(
                [edges.src['feat'], edges.dst['feat'], squared_distance],
                dim=-1)
        elif self.edge_features and not self.pairwise_distances:
            z2 = torch.cat(
                [edges.src['feat'], edges.dst['feat'], edges.data['feat']],
                dim=-1)
        else:
            z2 = torch.cat([edges.src['feat'], edges.dst['feat']], dim=-1)
        return {"e": self.pretrans(z2)}


class PNAGNN(nn.Module):
    """
    Principal Neighbourhood Aggregation Graph Neural Network [1]. This defines the message passing layers of the PNA model.

    Parameters
    ----------
    hidden_dim : int
        Dimension of the hidden layers.
    aggregators : List[str]
        List of aggregator functions to use.
    scalers : List[str]
        List of scaler functions to use. Options are "identity", "amplification", "attenuation".
    residual : bool, optional, default=True
        Whether to use residual connections.
    pairwise_distances : bool, optional, default=False
        Whether to use pairwise distances.
    activation : Union[Callable, str], optional, default="relu"
        Activation function to use.
    batch_norm : bool, optional, default=True
        Whether to use batch normalization in the layers before the aggregator.
    batch_norm_momentum : float, optional, default=0.1
        Momentum for the batch normalization layers.
    propagation_depth : int, optional, default=5
        Number of propagation layers.
    dropout : float, optional, default=0.0
        Dropout rate.
    posttrans_layers : int, optional, default=1
        Number of post-transformation layers.
    pretrans_layers : int, optional, default=1
        Number of pre-transformation layers.

    References
    ----------
    .. [1] Corso, G., Cavalleri, L., Beaini, D., Liò, P. & Veličković, P. Principal Neighbourhood Aggregation for Graph Nets. Preprint at https://doi.org/10.48550/arXiv.2004.05718 (2020).

    Examples
    --------
    >>> import numpy as np
    >>> from deepchem.feat.molecule_featurizers.conformer_featurizer import RDKitConformerFeaturizer
    >>> from deepchem.feat.graph_data import BatchGraphData
    >>> from deepchem.models.torch_models.pna_gnn import PNAGNN
    >>> featurizer = RDKitConformerFeaturizer(num_conformers=2)
    >>> smiles = ['C1=CC=NC=C1', 'CC(=O)C', 'C']
    >>> featurizer = RDKitConformerFeaturizer(num_conformers=2, rmsd_cutoff=1)
    >>> data = featurizer.featurize(smiles)
    >>> features = BatchGraphData(np.concatenate(data))
    >>> features = features.to_dgl_graph()
    >>> model = PNAGNN(hidden_dim=16,
    ...                aggregators=['mean', 'sum'],
    ...                scalers=['identity'])
    >>> output = model(features)
    """

    def __init__(self,
                 hidden_dim,
                 aggregators: List[str],
                 scalers: List[str],
                 residual: bool = True,
                 pairwise_distances: bool = False,
                 activation: Union[Callable, str] = "relu",
                 batch_norm: bool = True,
                 batch_norm_momentum=0.1,
                 propagation_depth: int = 5,
                 dropout: float = 0.0,
                 posttrans_layers: int = 1,
                 pretrans_layers: int = 1,
                 **kwargs):
        super(PNAGNN, self).__init__()

        self.mp_layers = nn.ModuleList()

        for _ in range(propagation_depth):
            self.mp_layers.append(
                PNALayer(in_dim=hidden_dim,
                         out_dim=int(hidden_dim),
                         in_dim_edges=hidden_dim,
                         aggregators=aggregators,
                         scalers=scalers,
                         pairwise_distances=pairwise_distances,
                         residual=residual,
                         dropout=dropout,
                         activation=activation,
                         avg_d={"log": 1.0},
                         posttrans_layers=posttrans_layers,
                         pretrans_layers=pretrans_layers,
                         batch_norm=batch_norm,
                         batch_norm_momentum=batch_norm_momentum),)
        self.atom_encoder = AtomEncoder(emb_dim=hidden_dim)
        self.bond_encoder = BondEncoder(emb_dim=hidden_dim)

    def forward(self, input_graph: dgl.DGLGraph):
        """
        Forward pass of the PNAGNN model.

        Parameters
        ----------
        input_graph : dgl.DGLGraph
            Input graph with node and edge features.

        Returns
        -------
        graph : dgl.DGLGraph
            Output graph with updated node features after applying the message passing layers.
        """
        graph = copy.deepcopy(input_graph)
        graph.ndata['feat'] = self.atom_encoder(graph.ndata['x'])
        graph.edata['feat'] = self.bond_encoder(graph.edata['edge_attr'])

        for mp_layer in self.mp_layers:
            graph.ndata['feat'] = mp_layer(graph)

        return graph


class PNA(nn.Module):
    """
    Message passing neural network for graph representation learning [1]_.

    Parameters
    ----------
    hidden_dim : int
        Hidden dimension size.
    target_dim : int
        Dimensionality of the output, for example for binary classification target_dim = 1.
    aggregators : List[str]
        Type of message passing functions. Options are 'mean','sum','max','min','std','var','moment3','moment4','moment5'.
    scalers : List[str]
        Type of normalization layers in the message passing network. Options are 'identity','amplification','attenuation'.
    readout_aggregators : List[str]
        Type of aggregators in the readout network.
    readout_hidden_dim : int, default None
       The dimension of the hidden layer in the readout network. If not provided, the readout has the same dimensionality of the final layer of the PNA layer, which is the hidden dimension size.
    readout_layers : int, default 1
        The number of linear layers in the readout network.
    residual : bool, default True
        Whether to use residual connections.
    pairwise_distances : bool, default False
        Whether to use pairwise distances.
    activation : Union[Callable, str]
        Activation function to use.
    batch_norm : bool, default True
        Whether to use batch normalization in the layers before the aggregator..
    batch_norm_momentum : float, default 0.1
        Momentum for the batch normalization layers.
    propagation_depth : int, default
        Number of propagation layers.
    dropout : float, default 0.0
        Dropout probability in the message passing layers.
    posttrans_layers : int, default 1
        Number of post-transformation layers.
    pretrans_layers : int, default 1
        Number of pre-transformation layers.

    References
    ----------
    .. [1] Corso, G., Cavalleri, L., Beaini, D., Liò, P. & Veličković, P. Principal Neighbourhood Aggregation for Graph Nets. Preprint at https://doi.org/10.48550/arXiv.2004.05718 (2020).

    Examples
    --------
    >>> import numpy as np
    >>> from deepchem.feat.graph_data import BatchGraphData
    >>> from deepchem.models.torch_models.pna_gnn import PNA
    >>> from deepchem.feat.molecule_featurizers.conformer_featurizer import RDKitConformerFeaturizer
    >>> smiles = ["C1=CC=CN=C1", "C1CCC1"]
    >>> featurizer = RDKitConformerFeaturizer(num_conformers=2)
    >>> data = featurizer.featurize(smiles)
    >>> features = BatchGraphData(np.concatenate(data))
    >>> features = features.to_dgl_graph()
    >>> target_dim = 1
    >>> model = PNA(hidden_dim=16, target_dim=target_dim)
    >>> output = model(features)
    >>> print(output.shape)
    torch.Size([1, 1])
    """

    def __init__(self,
                 hidden_dim: int,
                 target_dim: int,
                 aggregators: List[str] = ['mean'],
                 scalers: List[str] = ['identity'],
                 readout_aggregators: List[str] = ['mean'],
                 readout_hidden_dim: int = 1,
                 readout_layers: int = 2,
                 residual: bool = True,
                 pairwise_distances: bool = False,
                 activation: Union[Callable, str] = "relu",
                 batch_norm: bool = True,
                 batch_norm_momentum: float = 0.1,
                 propagation_depth: int = 5,
                 dropout: float = 0.0,
                 posttrans_layers: int = 1,
                 pretrans_layers: int = 1,
                 **kwargs):
        super(PNA, self).__init__()
        self.node_gnn = PNAGNN(hidden_dim=hidden_dim,
                               aggregators=aggregators,
                               scalers=scalers,
                               residual=residual,
                               pairwise_distances=pairwise_distances,
                               activation=activation,
                               batch_norm=batch_norm,
                               batch_norm_momentum=batch_norm_momentum,
                               propagation_depth=propagation_depth,
                               dropout=dropout,
                               posttrans_layers=posttrans_layers,
                               pretrans_layers=pretrans_layers)
        if readout_hidden_dim == 1:
            readout_hidden_dim = hidden_dim
        self.readout_aggregators = readout_aggregators
        self.output = MultilayerPerceptron(
            d_input=hidden_dim * len(self.readout_aggregators),
            d_hidden=(readout_hidden_dim,) * readout_layers,
            batch_norm=False,
            d_output=target_dim)

    def forward(self, graph: dgl.DGLGraph):
        graph = self.node_gnn(graph)
        readouts_to_cat = [
            dgl.readout_nodes(graph, 'feat', op=aggr)
            for aggr in self.readout_aggregators
        ]
        readout = torch.cat(readouts_to_cat, dim=-1)
        return self.output(readout)



"""Utility functions for working with PyTorch."""

import torch
from typing import Callable, Union, List


def get_activation(fn: Union[Callable, str]):
    """Get a PyTorch activation function, specified either directly or as a string.

    This function simplifies allowing users to specify activation functions by name.
    If a function is provided, it is simply returned unchanged.  If a string is provided,
    the corresponding function in torch.nn.functional is returned.
    """
    if isinstance(fn, str):
        return getattr(torch.nn.functional, fn)
    return fn


def unsorted_segment_sum(data: torch.Tensor, segment_ids: torch.Tensor,
                         num_segments: int) -> torch.Tensor:
    """Computes the sum along segments of a tensor. Analogous to tf.unsorted_segment_sum.

    Parameters
    ----------
    data: torch.Tensor
        A tensor whose segments are to be summed.
    segment_ids: torch.Tensor
        The segment indices tensor.
    num_segments: int
        The number of segments.

    Returns
    -------
    tensor: torch.Tensor

    Examples
    --------
    >>> segment_ids = torch.Tensor([0, 1, 0]).to(torch.int64)
    >>> data = torch.Tensor([[1, 2, 3, 4], [5, 6, 7, 8], [4, 3, 2, 1]])
    >>> num_segments = 2
    >>> result = unsorted_segment_sum(data=data,
                                  segment_ids=segment_ids,
                                  num_segments=num_segments)
    >>> result
    tensor([[5., 5., 5., 5.],
        [5., 6., 7., 8.]])

    """
    # length of segment_ids.shape should be 1
    assert len(segment_ids.shape) == 1

    # Shape of segment_ids should be equal to first dimension of data
    assert segment_ids.shape[-1] == data.shape[0]

    s = torch.prod(torch.tensor(data.shape[1:])).long()
    segment_ids = segment_ids.repeat_interleave(s).view(segment_ids.shape[0],
                                                        *data.shape[1:])

    # data.shape and segment_ids.shape should be equal
    assert data.shape == segment_ids.shape
    shape: List[int] = [num_segments] + list(data.shape[1:])
    tensor: torch.Tensor = torch.zeros(*shape).scatter_add(
        0, segment_ids, data.float())
    tensor = tensor.type(data.dtype)
    return tensor
